from __future__ import annotations

from typing import NamedTuple

from dissect.cstruct import cstruct

cim_def = """
// Mapping

struct mapping_header {
    uint32  signature;
    uint32  version;
    uint32  first_id;
    uint32  second_id;
    uint32  physical_page_count;
    uint32  mapping_entry_count;
};

struct mapping_header_xp {
    uint32  signature;
    uint32  version;
    uint32  physical_page_count;
    uint32  mapping_entry_count;
};

struct mapping_entry {
    uint32  page_number;
    uint32  page_crc;
    uint32  free_space;
    uint32  used_space;
    uint32  first_id;
    uint32  second_id;
};

struct mapping {
    uint32  signature;
    uint32  version;
    uint32  first_id;
    uint32  second_id;
    uint32  physical_page_count;
    uint32  mapping_entry_count;
    mapping_entry entries[mapping_entry_count];
    uint32  free_dword_count;
    char    free[free_dword_count * 4];
};

struct mapping_xp {
    uint32  signature;
    uint32  version;
    uint32  physical_page_count;
    uint32  mapping_entry_count;
    uint32  entries[mapping_entry_count];
    uint32  free_dword_count;
    char    free[free_dword_count * 4];
    uint32  footer_signature;
};

// Objects

struct toc_entry {
    uint32  record_id;
    uint32  offset;
    uint32  size;
    uint32  crc;
};

// Index page

struct index_page {
    uint32  signature;
    uint32  logical_id;
    uint32  _pad;
    uint32  root_page;
    uint32  record_count;
    uint32  unk0[record_count];
    uint32  children[record_count + 1];
    uint16  keys[record_count];
    uint16  string_definition_table_size;
    uint16  string_definition_table[string_definition_table_size];
    uint16  string_table_size;
    uint16  string_table[string_table_size + 1];
};

// Classes

enum CIM_TYPES : uint8 {
    INT16 = 0x2,
    INT32 = 0x3,
    REAL32 = 0x4,
    REAL64 = 0x5,
    STRING = 0x8,
    BOOLEAN = 0xB,
    OBJECT = 0xD,
    INT8 = 0x10,
    UINT8 = 0x11,
    UINT16 = 0x12,
    UINT32 = 0x13,
    INT64 = 0x14,
    UINT64 = 0x15,
    DATETIME = 0x65,
    REFERENCE = 0x66,
    CHAR16 = 0x67
};

enum ARRAY_STATES : uint8 {
    NOT_ARRAY = 0x0,
    ARRAY = 0x20
};

enum BOOLEAN_STATES : uint16 {
    FALSE = 0x0,
    TRUE = 0xFFFF
};

enum DYNPROPS_STATES : uint8 {
    NO_DYNPROPS = 0x1,
    HAS_DYNPROPS = 0x2
};

enum BUILTIN_QUALIFIERS {
    PROP_QUALIFIER_KEY = 0x1,
    PROP_QUALIFIER_READ_ACCESS = 0x3,
    CLASS_QUALIFIER_PROVIDER = 0x6,
    CLASS_QUALIFIER_DYNAMIC = 0x7,
    PROP_QUALIFIER_TYPE = 0xA
};

enum BUILTIN_PROPERTIES {
    PRIMARY_KEY = 0x1,
    READ = 0x2,
    WRITE = 0x3,
    VOLATILE = 0x4,
    PROVIDER = 0x6,
    DYNAMIC = 0x7,
    TYPE = 0xA
};

struct cim_type {
    CIM_TYPES       type;
    ARRAY_STATES    array_state;
    uint16          unk;
};

struct property_reference {
    uint32  name_offset;
    uint32  property_offset;
};

struct class_definition_property {
    cim_type    type;
    uint16      index;
    uint32      offset;
    uint32      level;
};

struct qualifier_reference {
    uint32      key_reference;
    uint8       unk;
    cim_type    type;
    char        data[0];
};

struct class_name_record {
    uint32  size;
    char    data[size - 4];
};

struct class_definition_header {
    uint32  super_class_name_size;
    wchar   super_class_name[super_class_name_size];
    uint64  timestamp;
    uint32  data_len;
    uint8   unk0;
    uint32  class_name_offset;
    uint32  default_values_metadata_size;
    class_name_record   class_name;
};

struct class_instance_header {
    wchar   name_hash[0x40];
    uint64  ts1;
    uint64  ts2;
    uint32  data_size;
    uint32  class_name_offset;
    uint8   unk0;
    char    remaining_data[data_size - 9];
};

struct class_instance_xp_header {
    wchar   name_hash[0x20];
    uint64  ts1;
    uint64  ts2;
    uint32  data_size;
    uint32  class_name_offset;
    uint8   unk0;
    char    remaining_data[data_size - 9];
};
"""

c_cim = cstruct().load(cim_def)

DATA_PAGE_SIZE = 0x2000

INDEX_PAGE_SIZE = 0x2000

INDEX_PAGE_INVALID = 0xFFFFFFFF
INDEX_PAGE_INVALID2 = 0x00000000

MAPPING_PAGE_ID_MASK = 0x3FFFFFFF
MAPPING_PAGE_UNAVAIL = 0xFFFFFFFF
MAPPING_FILE_CLEAN = 0x1
UNMAPPED_PAGE_VALUE = 0x3FFFFFFF

ROOT_NAMESPACE_NAME = "root"
SYSTEM_NAMESPACE_NAME = "__SystemClass"
NAMESPACE_CLASS_NAME = "__namespace"

ARRAY_STATES = c_cim.ARRAY_STATES
BOOLEAN_STATES = c_cim.BOOLEAN_STATES
DYNPROPS_STATES = c_cim.DYNPROPS_STATES

BUILTIN_QUALIFIERS = c_cim.BUILTIN_QUALIFIERS
BUILTIN_PROPERTIES = c_cim.BUILTIN_PROPERTIES

CIM_TYPES = c_cim.CIM_TYPES
CIM_TYPES_MAP = {
    CIM_TYPES.INT16: c_cim.int16,
    CIM_TYPES.INT32: c_cim.int32,
    CIM_TYPES.REAL32: c_cim.float,
    CIM_TYPES.REAL64: c_cim.double,
    CIM_TYPES.STRING: c_cim.uint32,
    CIM_TYPES.BOOLEAN: c_cim.BOOLEAN_STATES,
    CIM_TYPES.OBJECT: c_cim.uint32,
    CIM_TYPES.INT8: c_cim.int8,
    CIM_TYPES.UINT8: c_cim.uint8,
    CIM_TYPES.UINT16: c_cim.uint16,
    CIM_TYPES.UINT32: c_cim.uint32,
    CIM_TYPES.INT64: c_cim.int64,
    CIM_TYPES.UINT64: c_cim.uint64,
    CIM_TYPES.DATETIME: c_cim.uint32,
    CIM_TYPES.REFERENCE: c_cim.uint32,
    CIM_TYPES.CHAR16: c_cim.wchar,
}

CimType = int | float | str | bool


class ClassDefinitionPropertyState(NamedTuple):
    is_inherited: bool
    has_default_value: bool


class ClassInstancePropertyState(NamedTuple):
    use_default_value: bool
    is_initialized: bool



from __future__ import annotations

from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, BinaryIO

from dissect.cim.c_cim import MAPPING_PAGE_ID_MASK, UNMAPPED_PAGE_VALUE
from dissect.cim.exceptions import Error, UnmappedPageError

if TYPE_CHECKING:
    from dissect.cim.c_cim import c_cim
    from dissect.cim.cim import CIM


class Mapping:
    """Provides forward and reverse lookup of physical and logical pages."""

    def __init__(self, cim: CIM, fh: BinaryIO):
        self.cim = cim
        self.mapping = cim.map_type(fh)

        if self.mapping.signature != 0xABCD:
            raise Error("Invalid mapping file!")

        self.get_entry = lru_cache(256)(self.get_entry)

    def __getitem__(self, k: int) -> MappingEntry:
        if not isinstance(k, int):
            raise TypeError("Invalid type")
        return self.get_entry(k)

    @cached_property
    def _reverse_map(self) -> dict[int, int]:
        reverse_map = {}
        for i in range(self.mapping.mapping_entry_count):
            pnum = self.get_entry(i).page_number

            if pnum == UNMAPPED_PAGE_VALUE:
                continue

            reverse_map[pnum] = i

        return reverse_map

    def get_entry(self, logical_num: int) -> MappingEntry:
        if logical_num > self.mapping.mapping_entry_count:
            raise IndexError(logical_num)

        try:
            entry = self.mapping.entries[logical_num]
        except Exception:
            raise UnmappedPageError(logical_num)

        return MappingEntry(entry)

    def reverse(self, physical_num: int) -> int:
        if physical_num in self._reverse_map:
            return self._reverse_map[physical_num]

        raise UnmappedPageError(physical_num)


class MappingEntry:
    def __init__(self, entry: int | c_cim.mapping_entry):
        self.entry = entry

        if isinstance(entry, int):
            self.page_number = entry & MAPPING_PAGE_ID_MASK
            self.page_crc = None
            self.free_space = None
            self.used_space = None
            self.first_id = None
            self.second_id = None
        else:
            self.page_number = entry.page_number & MAPPING_PAGE_ID_MASK
            self.page_crc = entry.page_crc
            self.free_space = entry.free_space
            self.used_space = entry.used_space
            self.first_id = entry.first_id
            self.second_id = entry.second_id

    @property
    def is_mapped(self) -> bool:
        return self.page_number != UNMAPPED_PAGE_VALUE



from __future__ import annotations


class Error(Exception):
    pass


class ReferenceNotFoundError(Error):
    pass


class InvalidDatabaseError(Error):
    pass


class UnmappedPageError(Error):
    pass



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

from dissect.cim.c_cim import DATA_PAGE_SIZE, c_cim
from dissect.cim.exceptions import Error

if TYPE_CHECKING:
    from dissect.cim.cim import CIM
    from dissect.cim.index import Key
    from dissect.cim.mappings import Mapping


class Objects:
    def __init__(self, cim: CIM, fh: BinaryIO, mapping: Mapping):
        self.store = Store(cim, fh, mapping)

    def get(self, key: Key) -> BytesIO:
        if not key.is_data_reference:
            raise ValueError(f"Key is not a data reference: {key}")

        page = self.store.page(key.data_page)

        buf = page.data(key)
        data_len = key.data_length
        cur_len = len(buf)

        if cur_len == data_len:
            return BytesIO(buf)

        read_list = [buf]
        cur_page = key.data_page + 1
        while cur_len < data_len:
            next_buf = self.store.logical_page(cur_page)
            if cur_len + len(next_buf) > data_len:
                # Last page containing partial data
                chunk_size = data_len - cur_len
                read_list.append(next_buf[:chunk_size])
                cur_len += chunk_size
            else:
                # Entire page is data
                read_list.append(next_buf)
                cur_len += len(next_buf)

            cur_page += 1

        return BytesIO(b"".join(read_list))


class Store:
    def __init__(self, cim: CIM, fh: BinaryIO, mapping: Mapping):
        self.cim = cim
        self.fh = fh
        self.map = mapping

    def page(self, logical_num: int) -> DataPage:
        page_number = self.map.get_entry(logical_num).page_number
        buf = self.physical_page(page_number)
        return DataPage(self, BytesIO(buf), logical_num, page_number)

    def logical_page(self, logical_num: int) -> bytes:
        return self.physical_page(self.map.get_entry(logical_num).page_number)

    def physical_page(self, page_number: int) -> bytes:
        self.fh.seek(DATA_PAGE_SIZE * page_number)
        return self.fh.read(DATA_PAGE_SIZE)


class DataPage:
    def __init__(self, store: Store, fh: BinaryIO, logical_num: int, page_number: int):
        self.store = store
        self.fh = fh
        self.logical_num = logical_num
        self.page_number = page_number

        self.toc = TOC(fh)

    def data(self, key: Key) -> bytes:
        target_id = key.data_id
        target_size = key.data_length

        for i in range(self.toc.count):
            entry = self.toc[i]
            if entry.record_id == target_id:
                if entry.size < target_size:
                    raise Error(f"Entry size is smaller than target size: {entry.size} < {target_size}")

                if entry.size > DATA_PAGE_SIZE - entry.offset:
                    pass

                self.fh.seek(entry.offset)
                return self.fh.read(entry.size)

        raise IndexError

    def objects(self) -> None:
        pass


class TOC:
    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.entries = []

        while True:
            e = c_cim.toc_entry(self.fh)
            if e.record_id == 0 and e.offset == 0 and e.size == 0 and e.crc == 0:
                break
            self.entries.append(e)

        self.count = len(self.entries)

    def __getitem__(self, k: int) -> c_cim.toc_entry:
        return self.entries[k]



from __future__ import annotations

from dissect.cim.cim import CIM
from dissect.cim.exceptions import (
    Error,
    InvalidDatabaseError,
    ReferenceNotFoundError,
    UnmappedPageError,
)

__all__ = [
    "CIM",
    "Error",
    "InvalidDatabaseError",
    "ReferenceNotFoundError",
    "UnmappedPageError",
]



# Heavily inspired from https://github.com/fireeye/flare-wmi
#
# Information about e.g. data structures can also be found in:
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wmio/b44d0581-5bd3-40fc-95d7-01c1b1239820

from __future__ import annotations

from functools import cached_property
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cim.c_cim import (
    ARRAY_STATES,
    NAMESPACE_CLASS_NAME,
    ROOT_NAMESPACE_NAME,
    SYSTEM_NAMESPACE_NAME,
    c_cim,
)
from dissect.cim.classes import (
    ClassDefinition,
    ClassInstance,
    PropertyDefaultValues,
)
from dissect.cim.exceptions import Error, InvalidDatabaseError
from dissect.cim.index import Index, Key
from dissect.cim.mappings import Mapping
from dissect.cim.objects import Objects
from dissect.cim.utils import find_current_mapping, parse_object_path

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.cstruct import BaseType

    from dissect.cim.c_cim import (
        CimType,
    )
    from dissect.cim.classes import (
        ClassDefinitionProperty,
        ClassInstanceProperty,
        InstanceKey,
        Qualifier,
    )


class CIM:
    """Common Information Model."""

    def __init__(self, index: BinaryIO, objects: BinaryIO, mappings: list[BinaryIO]):
        self._findex = index
        self._fobjects = objects
        self._fmappings = mappings

        if not len(mappings):
            raise Error("No mappings")

        self.type_xp, self._current_map = find_current_mapping(self._fmappings)

        if not self._current_map:
            raise Error("Couldn't find current map file")

        if self.type_xp:
            self._map_header = c_cim.mapping_header_xp
            self._map_type = c_cim.mapping_xp
        else:
            self._map_header = c_cim.mapping_header
            self._map_type = c_cim.mapping
        self.map_type = self._map_type

        self.objects = Objects(self, objects, Mapping(self, self._current_map))

        # In some Windows 10 mappings, the index mapping doesn't seem to have a footer signature (0xdcba)
        # However, the objects mapping still has it, so we just read it here and check it
        footer_signature = c_cim.uint32(self._current_map)
        if footer_signature != 0xDCBA:
            raise Error("Invalid footer signature in objects mapping")

        try:
            self.index = Index(self, index, Mapping(self, self._current_map))
        except EOFError:
            raise InvalidDatabaseError("Invalid CIM database, possibly corrupt")

        self.root = self.namespace(ROOT_NAMESPACE_NAME)
        self.system = self.namespace(SYSTEM_NAMESPACE_NAME)

    @classmethod
    def from_directory(cls, path: Path | str) -> CIM:
        if not isinstance(path, Path):
            path = Path(path)

        path = path.resolve()
        if not path.is_dir():
            raise ValueError(f"Path {path} is not a directory")

        findex = path.joinpath("INDEX.BTR").open("rb")
        fobjects = path.joinpath("OBJECTS.DATA").open("rb")
        fmappings = [path.joinpath(f"MAPPING{i}.MAP").open("rb") for i in range(1, 4)]

        return cls(findex, fobjects, fmappings)

    def key(self, *args) -> Key:
        return Key(self, *args)

    def query(self, path: str, ns: Namespace | None = None) -> Namespace | Class | Instance:
        if ns is not None and not isinstance(ns, Namespace):
            raise TypeError("namespace should be an instance of Namespace")
        object_path = parse_object_path(path, ns)

        if object_path.hostname != "localhost":
            raise ValueError(f"Unsupported hostname: {object_path.hostname}")

        if not object_path.namespace:
            raise ValueError(f"Invalid ObjectPath: {object_path}")

        obj = self.namespace(object_path.namespace)

        if object_path.class_:
            obj = obj.class_(object_path.class_)

        if object_path.instance:
            obj = obj.instance(object_path.instance)

        return obj

    def namespace(self, name: str) -> Namespace:
        return Namespace(self, name)

    def _parse_instance(self, class_: Class, buf: BytesIO) -> ClassInstance:
        return ClassInstance(self, class_, buf)

    def get_class_definition(self, q: Key) -> ClassDefinition:
        if not q.reference():
            q = self.key().NS(SYSTEM_NAMESPACE_NAME).CD(q["CD"])
        return ClassDefinition(self, q.object())

    def get_class_instance(self, class_: Class, q: Key) -> ClassInstance:
        return self._parse_instance(class_, q.object())


class Namespace:
    def __init__(self, cim: CIM, name: str, class_instance: ClassInstance | None = None):
        self.cim = cim
        self.name = name
        self.class_instance = class_instance

    def __repr__(self) -> str:
        return f"<Namespace {self.name}>"

    def query(self, path: str) -> Namespace | Class | Instance:
        return self.cim.query(path, self)

    @property
    def ci(self) -> ClassInstance | None:
        return self.class_instance

    def parent(self) -> Namespace:
        raise NotImplementedError

    def class_(self, class_name: str) -> Class:
        q = self.cim.key().NS(self.name).CD(class_name)
        class_def = self.cim.get_class_definition(q)

        return Class(self.cim, self, class_def)

    @property
    def classes(self) -> Iterator[Class]:
        yielded = set()

        if self.name != SYSTEM_NAMESPACE_NAME:
            for class_ in self.cim.system.classes:
                class_.namespace = self
                if class_.name not in yielded:
                    yield class_
                    yielded.add(class_.name)

        q = self.cim.key().NS(self.name).CD()
        for ref in q.references():
            class_def = self.cim.get_class_definition(ref)
            class_ = Class(self.cim, self, class_def)

            if class_.name not in yielded:
                yield class_
                yielded.add(class_.name)

    def namespace(self, name: str) -> Namespace:
        main_name = f"{self.name}\\{name}".lower()
        for ns in self.namespaces:
            if ns.name.lower() == main_name:
                return ns

        raise IndexError

    @property
    def namespaces(self) -> Iterator[Namespace]:
        yielded = set()

        q = self.cim.key().NS(self.name).CI(NAMESPACE_CLASS_NAME).IL()
        class_def = self.cim.system.class_(NAMESPACE_CLASS_NAME)

        for ref in q.references():
            class_instance = self.cim.get_class_instance(class_def, ref)
            ns = Namespace(self.cim, f"{self.name}\\{class_instance.properties['Name'].value}", class_instance)

            if ns.name not in yielded:
                yield ns
                yielded.add(ns)

        if self.name == ROOT_NAMESPACE_NAME:
            yield self.cim.system  # Why do this?


class Class:
    def __init__(self, cim: CIM, namespace: Namespace, class_definition: ClassDefinition):
        self.cim = cim
        self.namespace = namespace
        self.class_definition = class_definition

        self._properties = None

    def __getattr__(self, attr: str) -> Any:
        try:
            return getattr(self.class_definition, attr)
        except AttributeError:
            return object.__getattribute__(self, attr)

    @property
    def name(self) -> str:
        return self.class_definition.class_name

    @property
    def ns(self) -> Namespace:
        return self.namespace

    @property
    def cd(self) -> ClassDefinition:
        return self.class_definition

    @property
    def derivation(self) -> list[Class]:
        """List from root to leaf of class layouts."""
        derivation = []

        class_ = self
        while True:
            derivation.append(class_)
            super_class_name = class_.class_definition.super_class_name
            if super_class_name == "":
                break

            class_ = self.namespace.class_(super_class_name)

        derivation.reverse()
        return derivation

    @cached_property
    def properties(self) -> dict[str, Property]:
        props = {}
        for class_ in self.derivation:
            for prop in class_.class_definition.properties.values():
                props[prop.name] = Property(self, prop)
        return props

    @property
    def property_default_values(self) -> PropertyDefaultValues:
        props = self.properties.values()
        props = sorted(props, key=lambda p: p.index)
        return PropertyDefaultValues(BytesIO(self.class_definition.default_values_data), props)

    @property
    def properties_length(self) -> int:
        off = 0
        for prop in self.properties.values():
            if prop.type.array_state == ARRAY_STATES.ARRAY:
                off += 0x4
            else:
                off += len(prop.ctype)
        return off

    def instance(self, key: str) -> Instance:
        for instance in self.instances:
            if instance.key == key:
                return instance

        raise IndexError

    @property
    def instances(self) -> Iterator[Instance]:
        yielded = set()

        q = self.cim.key().NS(self.namespace.name).CI(self.name).IL()

        for ref in q.references():
            class_instance = self.cim.get_class_instance(self, ref)
            instance = Instance(self.cim, self.namespace, self, class_instance)
            ikey = str(instance.key)

            if ikey not in yielded:
                yield instance
                yielded.add(ikey)


class Instance:
    def __init__(self, cim: CIM, namespace: Namespace, class_: Class, class_instance: ClassInstance):
        self.cim = cim
        self.namespace = namespace
        self.class_ = class_
        self.class_definition = class_.class_definition
        self.class_instance = class_instance

    @property
    def key(self) -> InstanceKey:
        return self.class_instance.key

    @property
    def name(self) -> str:
        return self.class_instance.class_name

    @property
    def ns(self) -> Namespace:
        return self.namespace

    @property
    def cd(self) -> ClassDefinition:
        return self.class_definition

    @property
    def ci(self) -> ClassInstance:
        return self.class_instance

    @property
    def properties(self) -> dict[str, ClassInstanceProperty]:
        return self.class_instance.properties


class Property:
    def __init__(self, class_: Class, prop: ClassDefinitionProperty):
        self.class_ = class_
        self._prop = prop

    @property
    def type(self) -> c_cim.cim_type:
        return self._prop.type

    @property
    def ctype(self) -> BaseType:
        return self._prop.ctype

    @property
    def qualifiers(self) -> dict[str, Qualifier]:
        return self._prop.qualifiers

    @property
    def name(self) -> str:
        return self._prop.name

    @property
    def index(self) -> int:
        return self._prop.index

    @property
    def offset(self) -> int:
        return self._prop.offset

    @property
    def level(self) -> int:
        return self._prop.level

    @property
    def is_inherited(self) -> bool:
        return self.class_.property_default_values.state[self.index].is_inherited

    @property
    def has_default_value(self) -> bool:
        return self.class_.property_default_values.state[self.index].has_default_value

    @property
    def default_value(self) -> CimType | list[CimType]:
        if not self.has_default_value:
            raise ValueError("Property has no default value!")

        if not self.is_inherited:
            # then the data is stored nicely in the CD prop data section
            v = self.class_.property_default_values.default_values[self.index]
            return self.class_.class_definition.property_data.get_value(v, self.type)

        # we have to walk up the derivation path looking for the default value
        rderivation = self.class_.derivation[:]
        rderivation.reverse()

        for ancestor_cl in rderivation:
            defaults = ancestor_cl.property_default_values
            state = defaults.state[self.index]
            if not state.has_default_value:
                raise Error("Property with inherited default value has bad ancestor (no default value)")

            if state.is_inherited:
                # keep trucking! look further up the ancestry tree.
                continue

            # else, this must be where the default value is defined
            v = defaults.default_values[self.index]
            return ancestor_cl.class_definition.property_data.get_value(v, self.type)

        raise Error("Unable to find ancestor class with default value")
